(*
    This file is part of the NumericRep project -
    which provides conversion IEEE floats.
    
    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(* Conversion between single precision binary IEEE Floating point
   representation and ML's LargeReal. *)
(* IEEESingleFloat.fromLargeReal is not implemented *)
functor FIEEESingleFloat() :>
          IEEE_FLOAT
            where type binary = Word32.word =
  struct
    type binary = Word32.word

    (* Should be bits 0-22 *)
    (* FIXME: I believe the low bits are special bits (NaN etc.) *)
    val significandBits = Word32.-(0wx800000, 0w1)

    (* Some compilers support different sizes of constants,
       so we need to do like this. *)
    fun setBit bit =
          Word32.<<(0w1, bit)

    (* Should be bits 23-30 *)
    val exponentBits = foldl Word32.+ 0w0
                         [0wx800000,
                         0wx1000000,
                         0wx2000000,
                         0wx4000000,
                         0wx8000000,
                        0wx10000000,
                        0wx20000000,
                        setBit 0w30]
    
    (* Should be bit 31 *)
    val signBit = setBit 0w31

    val exponentBias = 127

    val exponentShift = 0w23

    val significandDivide = LargeReal.fromInt 0x800000

    val _ = if LargeReal.radix = 2 then
              () (* This is good *)
            else
              print "The functor FIEEESingleFloat does not like your ML compiler!"
  
    fun toLargeReal binRep =
          let
            val sign = if Word32.andb(signBit, binRep) = 0w0 then
                         1.0
                       else
                         ~1.0
            val exponent = Word32.>>(Word32.andb(exponentBits, binRep), exponentShift)
            val significand = Word32.andb(significandBits, binRep)
            val significandReal = LargeReal.fromInt (Word32.toInt significand)
            val significandFrac =
                  LargeReal./(significandReal, significandDivide)
            val unsignedMantissa =
                  LargeReal.+(1.0, significandFrac)
          in
            LargeReal.fromManExp
              {man = LargeReal.*(sign, unsignedMantissa),
               exp = (Word32.toInt exponent) - exponentBias}
          end

    fun fromNormalDecimal sign digits exp =
          let
            (* FIXME: Not implemented! *)
          in
            Word32.fromInt 0
          end

    fun fromLargeReal value =
          Word32.fromInt 0
          (* let
            (* FIXME: LargeReal.toDecimal not supported in MLTon *)
            val {kind, sign, digits, exp} = LargeReal.toDecimal value
          in
            case kind of
              IEEEReal.NAN _ =>
                (* FIXME: Not correct! *)
                (print "Error: Only normal reals are supported!";
                 Word32.fromInt 0)
            | IEEEReal.INF =>
                (* FIXME: Not correct! *)
                (print "Error: Infinite reals not supported!";
                 Word32.fromInt 0)
            | IEEEReal.ZERO =>
                Word32.fromInt 0
            | IEEEReal.NORMAL =>
                fromNormalDecimal sign digits exp
            | IEEEReal.SUBNORMAL =>
                fromNormalDecimal sign digits exp
          end *)
  end

structure IEEESingleFloat = FIEEESingleFloat()